from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, List

import polars as pl

from harley.utils import parse_into_expr, register_plugin, parse_version

if TYPE_CHECKING:
    from polars.type_aliases import IntoExpr

if parse_version(pl.__version__) < parse_version("0.20.16"):
    from polars.utils.udfs import _get_shared_lib_location

    lib: str | Path = _get_shared_lib_location(__file__)
else:
    lib = Path(__file__).parent


def single_space(expr: IntoExpr) -> IntoExpr:
    """
    Replaces all whitespace to a single space from a string, then trims leading and trailing spaces.
    """
    expr = parse_into_expr(expr)
    return register_plugin(
        args=[expr],
        symbol="single_space",
        is_elementwise=True,
        lib=lib,
    )


def remove_all_whitespace(expr: IntoExpr) -> IntoExpr:
    """
    Removes all whitespace from a string.
    """
    expr = parse_into_expr(expr)
    return register_plugin(
        args=[expr],
        symbol="remove_all_whitespace",
        is_elementwise=True,
        lib=lib,
    )


def remove_non_word_characters(expr: IntoExpr) -> IntoExpr:
    """
    Removes all non-word characters. "Word characters" are [\w\s], i.e. alphanumeric, whitespace, and underscore ("_").
    """
    expr = parse_into_expr(expr)
    return register_plugin(
        args=[expr],
        symbol="remove_non_word_characters",
        is_elementwise=True,
        lib=lib,
    )


def anti_trim(expr: List[IntoExpr]) -> IntoExpr:
    """
    Replaces all whitespace to a single space from a string, then trims leading and trailing spaces.
    """
    expr = parse_into_expr(expr)
    return register_plugin(
        args=[expr],
        symbol="anti_trim",
        is_elementwise=True,
        lib=lib,
    )



from harley.utils import PolarsFrame
from polars import LazyFrame, Object, Unknown, DataFrame, col
from typing import List, Dict
from polars import DataType
from typing import OrderedDict, Any


def column_to_list(df: PolarsFrame, column: str) -> List:
    """
    Takes a PolarsFrame and a column name, extracts the specified column as a list, and
    returns it.

    :param df: PolarsFrame
    :type df: PolarsFrame
    :param column: The name of the column in
    the Polars DataFrame (`df`) that you want to extract and convert into a Python list
    :type column: str
    :return: a list of values from the specified column in the PolarsFrame dataframe.
    """
    if isinstance(df := df.select(column), LazyFrame):
        df = df.collect()
    series = df.select(column).to_series()
    return series.to_list()


def nested_fields(schema: OrderedDict[str, DataType]) -> Dict[str, DataType]:
    return OrderedDict(
        [(field, d_type) for field, d_type in schema if d_type.is_nested()]
    )


def two_columns_to_dictionary(
    df: DataFrame,
    key_col_name: str,
    value_col_name: str,
    allow_duplicates_keys: bool = False,
) -> Dict[Any, Any]:
    """
    Converts two columns from a DataFrame into a dictionary,
    with one column as keys and the other as values,
    handling potential issues like duplicate keys and
    non-hashable types.

    :param df: A DataFrame containing the data you want to convert into a dictionary
    :type df: DataFrame
    :param key_col_name: The name of the column in the DataFrame that will be used as the key in the resulting dictionary.
    This column should contain unique values that will serve as the keys in the dictionary mapping.
    :type key_col_name: str
    :param value_col_name: The name of the column in the DataFrame that contains the
    values you want to map to the keys in the dictionary.
    :type value_col_name: str
    :param allow_duplicates_keys: A boolean flag that determines whether duplicate keys are
    allowed in the resulting dictionary. If set to `False` (default), the function will raise a
    `ValueError` if duplicate keys are found in the specified key column, defaults to False
    :type allow_duplicates_keys: bool (optional)
    :return: Converts the two specified columns from the DataFrame into a dictionary
    where the key is the value from the key column and the value is the value from the value column.
    """
    df = df.select(key_col_name, value_col_name)
    key_dtype = df.dtypes[0]
    if key_dtype.is_nested():
        raise ValueError(
            f"Column {key_col_name} is of type {df.dtypes[0]} will not return a hashable type.",
            f"Therefore {key_col_name} cannot provide an acceptable key",
        )
    if key_dtype in [Object, Unknown]:
        raise ValueError(
            f"Column {key_col_name} is of type {df.dtypes[0]} and may not return a hashable type.",
            f"Therefore {key_col_name} cannot provide an acceptable key.",
        )
    if not allow_duplicates_keys:
        if df.filter(col(key_col_name).is_duplicated()).height > 0:
            raise ValueError(
                "Duplicate records found in key column.",
                "Duplicate keys will be overwritten and de-duplicated unpredicatbly.",
                "To allow duplicate keys, set `allow_duplicate_keys` to True",
                "Some of the duplicates:",
                column_to_list(
                    df.filter((col(key_col_name).is_duplicated())), key_col_name
                )[:3],
            )
    df = df.to_dicts()
    return {row[key_col_name]: row[value_col_name] for row in df}



import polars as pl
from harley.dataframe_validator import (
    DataFrameMissingColumnError,
    validate_presence_of_columns,
    DataSchemaError,
    validate_schema,
    validate_absence_of_columns,
    DataFrameProhibitedColumnError,
)
import pytest
from typing import OrderedDict, Union, Dict
from harley.utils import polars_frames

age_name_data = {"age": [1, 2, 3], "name": ["jose", "li", "luisa"]}


@pytest.mark.parametrize("frame_type", polars_frames)
def test_validate_presence_of_columns_fail(frame_type: pl.DataType):
    source_df = frame_type(age_name_data)
    with pytest.raises(DataFrameMissingColumnError) as exec_info:
        validate_presence_of_columns(source_df, ["name", "age", "fun"])
    assert (
        exec_info.value.args[0]
        == "The ['fun'] columns are not included in the DataFrame with the following columns ['age', 'name']"
    )


@pytest.mark.parametrize("frame_type", polars_frames)
def test_validate_presence_of_columns_pass(frame_type: pl.DataType):
    source_df = frame_type(age_name_data)
    validate_presence_of_columns(source_df, ["name"])


def create_dict_parameters(frames=polars_frames, dict_types=[dict, OrderedDict]):
    parameter_params = []
    for frame in frames:
        for dict_type in dict_types:
            parameter_params.append((frame, dict_type))
    return parameter_params


pytest_dict_frame_parameters = create_dict_parameters()


@pytest.mark.parametrize("frame_type, dict_type", pytest_dict_frame_parameters)
def test_validate_schema_raises_when_struct_field_missing(
    frame_type: pl.DataType, dict_type: Union[Dict, OrderedDict]
):
    source_df = frame_type(age_name_data)
    required_schema = dict_type(
        [
            ("name", pl.Utf8),
            ("city", pl.Utf8),
        ]
    )
    with pytest.raises(DataSchemaError):
        validate_schema(source_df, required_schema)


@pytest.mark.parametrize("frame_type, dict_type", pytest_dict_frame_parameters)
def test_validate_schema_does_nothing_when_schema_matches(
    frame_type: pl.DataType, dict_type: Union[Dict, OrderedDict]
):
    df = frame_type(age_name_data)
    required_schema = dict_type(
        [
            ("name", pl.Utf8),
            ("age", pl.Int64),
        ]
    )
    validate_schema(df, required_schema)


@pytest.mark.parametrize("frame_type", polars_frames)
def test_validate_absence_of_columns_passes_when_no_match(frame_type: pl.DataType):
    df = frame_type(age_name_data)
    prohibited_columns = ["fun", "ip_address"]
    validate_absence_of_columns(df=df, prohibited_col_names=prohibited_columns)


@pytest.mark.parametrize("frame_type", polars_frames)
def test_validate_absence_of_columns_fails_when_match(frame_type: pl.DataType):
    df = frame_type(age_name_data)
    prohibited_columns = ["name"]
    with pytest.raises(DataFrameProhibitedColumnError):
        validate_absence_of_columns(df=df, prohibited_col_names=prohibited_columns)



from harley.schema_helpers import print_schema_as_code
from typing import OrderedDict
import polars.datatypes as T
import pytest
from io import StringIO
import sys


@pytest.mark.parametrize(
    "schema, prepend_pl, exp",
    [
        (
            OrderedDict([("col_1", T.String), ("col_2", T.Array)]),
            True,
            'OrderedDict([("col_1", pl.String),("col_2", pl.Array),])',
        ),
        (
            OrderedDict([("col_1", T.String), ("col_2", T.Array)]),
            False,
            'OrderedDict([("col_1", String),("col_2", Array),])',
        ),
        (
            OrderedDict([("col_1", T.String)]),
            False,
            'OrderedDict([("col_1", String),])',
        ),
        (
            OrderedDict([("col_1", T.String), ("col_2", T.Array), ("col_3", T.Int64)]),
            False,
            'OrderedDict([("col_1", String),("col_2", Array),("col_3", Int64),])',
        ),
    ],
)
def test_print_schema_as_code(schema: OrderedDict, prepend_pl: bool, exp: str):
    captured_output = StringIO()
    sys.stdout = captured_output
    print_schema_as_code(schema, prepend_pl)
    sys.stdout = sys.__stdout__
    res = captured_output.getvalue().rstrip("\n")
    assert res == exp



from typing import OrderedDict
from polars import DataType


def print_schema_as_code(
    schema: OrderedDict[str, DataType], prepend_pl: bool = False
) -> None:
    """
    Takes a schema as input and prints it as valid code, with an
    option to prepend 'pl.' to the column types.

    :param schema: An OrderedDict where the keys are column names (strings)
    and the values are data types (DataType)
    :type schema: OrderedDict[str, DataType]
    :param prepend_pl: The `prepend_pl` parameter is a boolean flag that determines whether to prepend
    'pl.' to the column types in the schema before printing it as code. If `prepend_pl` is set to True,
    each column type in the schema will be prefixed with 'pl.', defaults to False
    :type prepend_pl: bool (optional)
    """
    if prepend_pl:
        schema = OrderedDict(
            [(field, "pl." + str(d_type)) for field, d_type in schema.items()]
        )
    out = "OrderedDict(["
    for name, d_type in schema.items():
        out += '("' + name + '", ' + str(d_type) + "),"
    out += "])"
    print(out)



from harley.utils import PolarsFrame
from polars import LazyFrame
from typing import Dict, OrderedDict, Union, List


class DataFrameMissingColumnError(ValueError):
    """Raise this when there's a DataFrame column error."""


class DataSchemaError(ValueError):
    """Raise this when schema validation fails"""


class DataFrameProhibitedColumnError(ValueError):
    """Raise this when a DataFrame includes prohibited columns."""


def validate_presence_of_columns(
    df: PolarsFrame, required_col_names: List[str]
) -> None:
    """Validate the presence of column names in a DataFrame.

    :param df: A spark DataFrame.
    :type df: DataFrame`
    :param required_col_names: List of the required column names for the DataFrame.
    :type required_col_names: :py:class:`list` of :py:class:`str`
    :return: None.
    :raises DataFrameMissingColumnError: if any of the requested column names are
    not present in the DataFrame.
    """
    if isinstance(df, LazyFrame):
        all_col_names = df.collect_schema().names()
    else:
        all_col_names = df.columns
    missing_col_names = [x for x in required_col_names if x not in all_col_names]
    error_message = f"The {missing_col_names} columns are not included in the DataFrame with the following columns {all_col_names}"
    if missing_col_names:
        raise DataFrameMissingColumnError(error_message)


def validate_schema(df: PolarsFrame, required_schema: Union[Dict, OrderedDict]) -> None:
    """
    Validates whether a given DataFrame matches an expected schema.

    :param df: A PolarsFrame object representing a DataFrame or a LazyFrame
    :type df: PolarsFrame
    :param required_schema: The `required_schema` parameter is a dictionary or an ordered dictionary
    that represents the schema that the input `df` (PolarsFrame) should adhere to. The function
    `validate_schema` compares the schema of the input dataframe with the required schema and raises a
    `DataSchemaError` if they do not match.
    :type required_schema: Union[Dict, OrderedDict]
    """
    if isinstance(df, LazyFrame):
        df_schema = df.collect_schema()
    else:
        df_schema = df.schema
    if isinstance(required_schema, dict):
        df_schema = dict(df_schema)
    if not required_schema == df_schema:
        raise DataSchemaError("The Frame did not match the expected schema")


def validate_absence_of_columns(
    df: PolarsFrame, prohibited_col_names: List[str]
) -> None:
    """
    Checks for the presence of prohibited columns in a DataFrame and raises an error if any are found.

    :param df: A PolarsFrame object representing a DataFrame or LazyFrame
    :type df: PolarsFrame
    :param prohibited_col_names: A list of column names that should not be present in the DataFrame `df`. The function `validate_absence_of_columns` checks if
    Raises a `DataFrameProhibitedColumnError` if any of the columns are present.
    :type prohibited_col_names: List[str]
    """
    if isinstance(df, LazyFrame):
        df_cols = df.collect_schema().names()
    else:
        df_cols = df.columns
    present_prohibited_col_names = [
        col for col in prohibited_col_names if col in df_cols
    ]
    if present_prohibited_col_names:
        raise DataFrameProhibitedColumnError(
            "Prohibited columns present:", present_prohibited_col_names
        )



from __future__ import annotations
from polars import Expr
import polars as pl
from typing import Any, List
from pathlib import Path
from typing import TYPE_CHECKING, Union
from harley.utils import parse_into_expr, register_plugin, parse_version


def null_between(col: str, lower: str, upper: str) -> Expr:
    """
    The function `null_between` returns an expression to check if a column value falls within a range,
    handling null values appropriately.

    :param col: The column for which you want to check if the values are
    null or fall within a certain range defined by the `lower` and `upper` bounds.
    :type col: str
    :param lower: The lower bound column or value that you want to compare against.
    :type lower: str
    :param upper: The upper bound column or value that you want to compare against
    :type upper: str
    :return: The `null_between` function is returning a polars expression that checks if a column
    `col` falls between two other columns `lower` and `upper`, handling cases where any of the columns
    are null.
    """
    return (
        pl.when(pl.col(lower).is_null() & pl.col(upper).is_null())
        .then(False)
        .when(pl.col(col).is_null())
        .then(False)
        .when(pl.col(lower).is_null())
        .then(pl.col(col) <= pl.col(upper))
        .when(pl.col(upper).is_null())
        .then(pl.col(col) >= pl.col(lower))
        .otherwise(pl.col(col).is_between(pl.col(lower), pl.col(upper)))
    )


def multi_equals(cols: List[str], val: Any) -> Expr:
    """
    Creates a query that checks if multiple columns have a specified value.

    :param cols: List[str]
    :type cols: List[str]
    :param val: The value that you want to compare the columns against in the
    `multi_equals` function
    :type val: Any
    :return: The function `multi_equals` returns an expression that represents the logical AND operation
    of equality checks between the columns specified in the `cols` list and the value `val`.
    """
    query = [pl.col(name) == val for name in cols]
    return pl.all_horizontal(query)


if TYPE_CHECKING:
    from polars.type_aliases import IntoExpr
if parse_version(pl.__version__) < parse_version("0.20.16"):
    from polars.utils.udfs import _get_shared_lib_location

    lib: str | Path = _get_shared_lib_location(__file__)
else:
    lib = Path(__file__).parent


def approx_equal(
    col_1: IntoExpr, col_2: IntoExpr, threshold: Union[float, int]
) -> IntoExpr:
    """
    Compares two columns for approximate equality within a specified threshold.

    :param col_1: The first column or expression that you want to compare for approximate
    equality
    :type col_1: IntoExpr
    :param col_2: The second column or value that you want to
    compare for approximate equality with col_1
    :type col_2: IntoExpr
    :param threshold: Specifies the  maximum allowable difference between the values in `col_1` and `col_2` for them to be considered
    approximately equal.
    :type threshold: Union[float, int]
    :return: boolean typed expression.
    """
    col_1 = parse_into_expr(col_1)
    col_2 = parse_into_expr(col_2)
    return register_plugin(
        args=[col_1, col_2],
        symbol="approx_equal",
        is_elementwise=True,
        lib=lib,
        kwargs={"threshold": float(threshold)},
    )



from __future__ import annotations
from pathlib import Path
from typing import TYPE_CHECKING
import polars as pl
from harley.utils import parse_into_expr, register_plugin, parse_version

if TYPE_CHECKING:
    from polars.type_aliases import IntoExpr
if parse_version(pl.__version__) < parse_version("0.20.16"):
    from polars.utils.udfs import _get_shared_lib_location

    lib: str | Path = _get_shared_lib_location(__file__)
else:
    lib = Path(__file__).parent


def is_null_or_blank(expr: IntoExpr, all_white_space_as_null: bool = False) -> IntoExpr:
    """
    Returns True if null/blank/empty list.
    Otherwise, False.
    """
    expr = parse_into_expr(expr)
    return register_plugin(
        args=[expr],
        symbol="is_null_or_blank",
        is_elementwise=True,
        lib=lib,
        kwargs={"all_white_space_as_null": all_white_space_as_null},
    )


def is_falsey(expr: IntoExpr) -> IntoExpr:
    """
    Returns True if null/False, otherwise False.
    """
    expr = parse_into_expr(expr)
    return register_plugin(
        args=[expr],
        symbol="is_falsey",
        is_elementwise=True,
        lib=lib,
    )


def is_false(expr: IntoExpr) -> IntoExpr:
    """
    Returns True if False, otherwise False.
    Whitespace aware.
    """
    expr = parse_into_expr(expr)
    return register_plugin(
        args=[expr],
        symbol="is_false",
        is_elementwise=True,
        lib=lib,
    )


def is_truthy(expr: IntoExpr) -> IntoExpr:
    """
    Returns True if True, otherwise False.
    Whitespace aware.
    """
    expr = parse_into_expr(expr)
    return register_plugin(
        args=[expr],
        symbol="is_truthy",
        is_elementwise=True,
        lib=lib,
    )


def is_true(expr: IntoExpr) -> IntoExpr:
    """
    Returns True if True, otherwise False.
    Whitespace aware.
    """
    expr = parse_into_expr(expr)
    return register_plugin(
        args=[expr],
        symbol="is_true",
        is_elementwise=True,
        lib=lib,
    )



from __future__ import annotations

from pathlib import Path

import polars as pl

from harley.utils import parse_version
from harley.dataframe_helper import column_to_list
from harley.transformations import snake_case_column_names, flatten_struct
from harley.string_functions import (
    single_space,
    remove_all_whitespace,
    remove_non_word_characters,
    anti_trim,
)
from harley.to_boolean import is_null_or_blank
from harley.column_functions import approx_equal
from harley.schema_helpers import print_schema_as_code
from harley.maths import div_or_else


if parse_version(pl.__version__) < parse_version("0.20.16"):
    from polars.utils.udfs import _get_shared_lib_location

    lib: str | Path = _get_shared_lib_location(__file__)
else:
    lib = Path(__file__).parent

__all__ = [
    "column_to_list",
    "single_space",
    "snake_case_column_names",
    "remove_all_whitespace",
    "remove_non_word_characters",
    "anti_trim",
    "is_null_or_blank",
    "snake_case_column_names",
    "approx_equal",
    "print_schema_as_code",
    "flatten_struct",
    "div_or_else",
]



from harley.utils import PolarsFrame
from .harley import columns_to_snake_case  # noqa
from typing import Union, List, Dict
from polars import col, Struct, Expr, LazyFrame

import warnings


def snake_case_column_names(df: PolarsFrame) -> PolarsFrame:
    """
    Takes a PolarsFrame, converts its column names to snake case,
    and returns the modified PolarsFrame.
    
    :param df: A PolarsFrame object, which can be eith er a LazyFrame or DataFrame.
    :type df: PolarsFrame
    :return: The function `snake_case_column_names` is returning a PolarsFrame with column names
    converted to snake case.
    """
    if isinstance(df, LazyFrame):
        all_column_names = df.collect_schema().names()
    else:
        all_column_names = df.columns
    new_col_names = columns_to_snake_case(all_column_names)
    return df.rename(new_col_names)


class ColumnNameRepeatedError(ValueError):
    """raised when a column name would be repeated after flatten_struct"""


def flatten_struct(
    df: PolarsFrame,
    struct_columns: Union[str, List[str]],
    separator: str = ":",
    drop_original_struct: bool = True,
    recursive: bool = False,
    limit: int = None,
) -> PolarsFrame:
    """
    Takes a PolarsFrame and flattens specified struct columns into
    separate columns using a specified separator,
    with options to control recursion and limit the number
    of flattening levels.

    :param df: A PolarsFrame, either a LazyFrame or DataFrame.
    :type df: PolarsFrame
    :param struct_columns: The column or columns in the PolarsFrame that contain struct data.
    This function is designed to flatten the struct data into separate columns based on the fields within the struct.
    :type struct_columns: Union[str, List[str]]
    :param separator: Specifies the character or string that will be used to separate the original
    column name from the nested field names when flattening a nested struct column.
    :type separator: str (optional)
    :param drop_original_struct: Determines whether the original struct columns should be dropped after flattening or not,
    defaults to True.
    :type drop_original_struct: bool (optional)
    :param recursive: Determines whether the flattening process should be applied recursively to
    all levels of nested structures within the specified struct columns, defaults to False.
    :type recursive: bool (optional)
    :param limit: Determines the maximum number of levels to flatten the struct columns.
    If `limit` is set to a positive integer, the function will flatten the struct columns up to that specified level.
    If `limit` is set to `None`, there is no limit.
    :type limit: int
    :return: returns a PolarsFrame.
    """
    if isinstance(struct_columns, str):
        struct_columns = [struct_columns]
    if not recursive:
        limit = 1
    if limit is not None and not isinstance(limit, int):
        raise ValueError("limit must be a positive integer or None")
    if limit is not None and limit < 0:
        raise ValueError("limit must be a positive integer or None")
    if limit == 0:
        warnings.warn("limit of 0 will result in no transformations")
        return df
    ldf = df.lazy()  # noop if df is LazyFrame
    all_column_names = ldf.collect_schema().names()
    if any(separator in (witness := column) for column in all_column_names):
        warnings.warn(
            f'separator "{separator}" found in column names, e.g. "{witness}". '
            "If columns would be repeated, this function will error"
        )
    non_struct_columns = list(set(ldf.collect_schema().names()) - set(struct_columns))
    struct_schema = ldf.select(*struct_columns).collect_schema()
    col_dtype_expr_names = [(struct_schema[c], col(c), c) for c in struct_columns]
    result_names: Dict[str, Expr] = {}
    level = 0
    while (limit is None and col_dtype_expr_names) or (
        limit is not None and level < limit
    ):
        level += 1
        new_col_dtype_exprs = []
        for dtype, col_expr, name in col_dtype_expr_names:
            if not isinstance(dtype, Struct):
                if name in result_names:
                    raise ColumnNameRepeatedError(
                        f"Column name {name} would be created at least twice after flatten_struct"
                    )
                result_names[name] = col_expr
                continue
            if any(separator in (witness := field.name) for field in dtype.fields):
                warnings.warn(
                    f'separator "{separator}" found in field names, e.g. "{witness}" in {name}. '
                    "If columns would be repeated, this function will error"
                )
            new_col_dtype_exprs += [
                (
                    field.dtype,
                    col_expr.struct.field(field.name),
                    name + separator + field.name,
                )
                for field in dtype.fields
            ]
            if not drop_original_struct:
                ldf = ldf.with_columns(
                    col_expr.struct.field(field.name).alias(
                        name + separator + field.name
                    )
                    for field in dtype.fields
                )
        col_dtype_expr_names = new_col_dtype_exprs
    if drop_original_struct and level == limit and col_dtype_expr_names:
        for _, col_expr, name in col_dtype_expr_names:
            result_names[name] = col_expr
    if any((witness := column) in non_struct_columns for column in result_names):
        raise ColumnNameRepeatedError(
            f"Column name {witness} would be created after flatten_struct, but it's already a non-struct column"
        )
    if drop_original_struct:
        ldf = ldf.select(
            [col(c) for c in non_struct_columns]
            + [col_expr.alias(name) for name, col_expr in result_names.items()]
        )

    if isinstance(df, LazyFrame):
        return ldf
    return ldf.collect()



from harley.maths import div_or_else
import pytest
from harley.utils import polars_frames
from polars import DataFrame, LazyFrame
from polars.testing import assert_frame_equal
from typing import List, Dict, Any, Union


@pytest.mark.parametrize("frame_type", polars_frames)
@pytest.mark.parametrize(
    "data, exp",
    (
        ({"dividend": [1, 2, 3, None], "divisor": [1, 1, 0, None]}, [1, 2, 0, None]),
        (
            {"dividend": [1.5, 2.5, 3, None], "divisor": [1, 1, 0, None]},
            [1.5, 2.5, 0, None],
        ),
        (
            {"dividend": [0.5, 3, 3, None], "divisor": [0.5, 1.5, 0.0, None]},
            [1.0, 2.0, 0.0, None],
        ),
    ),
)
def test_div_or_else(
    frame_type: Union[DataFrame, LazyFrame], data: Dict[str, List[Any]], exp: List
):
    inp = frame_type(data)
    exp = DataFrame({"res": exp})
    if isinstance(
        res := inp.select(res=div_or_else(dividend="dividend", divisor="divisor")),
        LazyFrame,
    ):
        res = res.collect()
    assert_frame_equal(res, exp)


@pytest.mark.parametrize("frame_type", polars_frames)
def test_div_or_else_alt_or_else(frame_type: Union[DataFrame, LazyFrame]):
    inp = frame_type({"dividend": [1.5, 2.5, 3, None], "divisor": [1, 1, 0, None]})
    exp = DataFrame({"res": [1.5, 2.5, 5, None]})
    if isinstance(
        res := inp.select(
            res=div_or_else(dividend="dividend", divisor="divisor", or_else=5)
        ),
        LazyFrame,
    ):
        res = res.collect()
    assert_frame_equal(res, exp)



import pytest

import polars as pl
from polars.testing import assert_series_equal
from string import ascii_letters

import harley


@pytest.mark.parametrize("lazy", [True, False])
def test_single_space(lazy: bool):
    string_data = [
        ("  I\t\r\v\n go   ", "I go"),
        ("    to", "to"),
        ("school   ", "school"),
        ("by bus", "by bus"),
        (None, None),
    ]
    frame = pl.DataFrame([{"input": inp, "expected": exp} for inp, exp in string_data])
    if lazy:
        frame = frame.lazy()
    frame = frame.with_columns(actual=harley.single_space("input"))
    if lazy:
        frame = frame.collect()
    assert_series_equal(frame["actual"], frame["expected"], check_names=False)


@pytest.mark.parametrize("lazy", [True, False])
def test_remove_all_whitespace(lazy: bool):
    string_data = [
        ("  I\t\r\v\n go   ", "Igo"),
        ("    to", "to"),
        ("school   ", "school"),
        ("by bus", "bybus"),
        (None, None),
    ]
    frame = pl.DataFrame([{"input": inp, "expected": exp} for inp, exp in string_data])
    if lazy:
        frame = frame.lazy()
    frame = frame.with_columns(actual=harley.remove_all_whitespace("input"))
    if lazy:
        frame = frame.collect()
    assert_series_equal(frame["actual"], frame["expected"], check_names=False)


@pytest.mark.parametrize("lazy", [True, False])
def test_remove_non_space_characters(lazy: bool):
    string_data = [
        ("\t\r\v\n ", "\t\r\v\n "),
        (ascii_letters, ascii_letters),
        ("1234567890", "1234567890"),
        ("_", "_"),
        ("", ""),
        ("!@#$%^&*()", ""),
        ("!I @go #to $school_%by *bus.?", "I go to school_by bus"),
        (None, None),
    ]
    frame = pl.DataFrame([{"input": inp, "expected": exp} for inp, exp in string_data])
    if lazy:
        frame = frame.lazy()
    frame = frame.with_columns(actual=harley.remove_non_word_characters("input"))
    if lazy:
        frame = frame.collect()
    assert_series_equal(frame["actual"], frame["expected"], check_names=False)


@pytest.mark.parametrize("lazy", [True, False])
def test_anti_trim(lazy: bool):
    string_data = [
        ("  I\t\r\v\n go   ", "  Igo   "),
        ("\t\r\v\n to", "\t\r\v\n to"),
        ("school   ", "school   "),
        ("by bus", "bybus"),
        ("    ", "    "),
        ("", ""),
        (None, None),
    ]
    frame = pl.DataFrame([{"input": inp, "expected": exp} for inp, exp in string_data])
    if lazy:
        frame = frame.lazy()
    frame = frame.with_columns(actual=harley.anti_trim("input"))
    if lazy:
        frame = frame.collect()
    assert_series_equal(frame["actual"], frame["expected"], check_names=False)



from harley.column_functions import null_between
from harley.utils import polars_frames
import pytest
from polars import DataFrame, LazyFrame, DataType
from polars.testing import assert_frame_equal
from typing import Tuple, Any, Union, List
from harley.column_functions import multi_equals
from harley.column_functions import approx_equal


@pytest.mark.parametrize("frame_type", polars_frames)
@pytest.mark.parametrize(
    "inp, exp, thresh",
    (
        (
            [[1.1, 1.1, 1.02, 1.02], [1.05, 11.6, 1.09, 1.34]],
            [True, False, True, False],
            0.1,
        ),
        (
            [[12, 20, 44, 32], [14, 26, 41, 9]],
            [True, False, True, False],
            5,
        ),
        (
            [[12, 20, 44, 32], [14, 26, 41, 9]],
            [True, False, True, False],
            4.1,
        ),
        (
            [[12, 20, 44, 32], [14, 26, 43, 9]],
            [False, False, True, False],
            1.9,
        ),
    ),
)
def test_approx_equal(
    frame_type: DataType,
    inp: List[List[Union[float, int]]],
    exp: List[bool],
    thresh: float,
):
    schema = ["left", "right"]
    data = frame_type(dict(zip(schema, inp)))
    exp = DataFrame({"res": exp})
    if isinstance(
        res := data.select(
            res=approx_equal(col_1="left", col_2="right", threshold=thresh)
        ),
        LazyFrame,
    ):
        res = res.collect()
    assert_frame_equal(res, exp)


@pytest.mark.parametrize("frame_type", polars_frames)
@pytest.mark.parametrize(
    "inp, exp",
    (
        (({"lower_age": [17], "upper_age": [None], "age": [94]}), True),
        (({"lower_age": [17], "upper_age": [None], "age": [10]}), False),
        (({"lower_age": [None], "upper_age": [10], "age": [5]}), True),
        (({"lower_age": [None], "upper_age": [10], "age": [88]}), False),
        (({"lower_age": [10], "upper_age": [15], "age": [11]}), True),
        (({"lower_age": [None], "upper_age": [None], "age": [11]}), False),
        (({"lower_age": [3], "upper_age": [5], "age": [None]}), False),
        (({"lower_age": [3], "upper_age": [3], "age": [None]}), False),
        (({"lower_age": [3], "upper_age": [3], "age": [3]}), True),
        (({"lower_age": [3], "upper_age": [4], "age": [5]}), False),
        (({"lower_age": [None], "upper_age": [None], "age": [None]}), False),
    ),
)
def test_null_between(
    frame_type: Union[DataFrame, LazyFrame], inp: Tuple[Any], exp: bool
):
    df = frame_type(inp)
    exp = DataFrame({"res": [exp]})
    if isinstance(
        res := df.select(
            res=null_between(col="age", lower="lower_age", upper="upper_age")
        ),
        LazyFrame,
    ):
        res = res.collect()
    assert_frame_equal(exp, res)


@pytest.mark.parametrize("frame_type", polars_frames)
def test_multi_equals(frame_type: Union[DataFrame, LazyFrame]):
    data = frame_type(
        [
            {"col_1": "cat", "col_2": "cat"},
            {"col_1": "cat", "col_2": "dog"},
            {"col_1": "dog", "col_2": "cat"},
            {"col_1": "dog", "col_2": "dog"},
        ]
    )
    exp = DataFrame({"res": [True, False, False, False]})
    if isinstance(
        res := data.select(res=multi_equals(cols=["col_1", "col_2"], val="cat")),
        LazyFrame,
    ):
        res = res.collect()
    assert_frame_equal(res, exp)



from harley import column_to_list
from harley.dataframe_helper import nested_fields
from polars import (
    Decimal,
    Float32,
    Float64,
    Int8,
    Int16,
    Int32,
    Int64,
    UInt8,
    UInt32,
    UInt64,
    Date,
    Datetime,
    Duration,
    Time,
    Array,
    List,
    Struct,
    String,
    Categorical,
    Enum,
    Utf8,
    Binary,
    Boolean,
    Null,
    Object,
    Unknown,
    LazyFrame,
)
import pytest
from datetime import datetime
from harley.utils import polars_frames
from typing import OrderedDict, Union
from harley.dataframe_helper import two_columns_to_dictionary
from polars import DataFrame

floats = [4.0, 5.0, 6.0, 7.0, 8.0]
data = {
    "integer": [1, 2, 3, 4, 5],
    "date": [
        datetime(2022, 1, 1),
        datetime(2022, 1, 2),
        datetime(2022, 1, 3),
        datetime(2022, 1, 4),
        datetime(2022, 1, 5),
    ],
    "float": floats,
}


@pytest.mark.parametrize("frame_type", polars_frames)
def test_column_to_list(frame_type: Union[DataFrame, LazyFrame]):
    frame = frame_type(data)
    exp = floats
    res = column_to_list(frame, "float")
    assert exp == res


def test_nested_fields():
    all_column_types = [
        Array,
        List,
        Struct,
        Decimal,
        Float32,
        Float64,
        Int8,
        Int16,
        Int32,
        Int64,
        UInt8,
        UInt32,
        UInt64,
        Date,
        Datetime,
        Duration,
        Time,
        String,
        Categorical,
        Enum,
        Utf8,
        Binary,
        Boolean,
        Null,
        Object,
        Unknown,
    ]
    schema = [
        ("col_" + str(i), col_type) for i, col_type in enumerate(all_column_types)
    ]
    res = nested_fields(schema=schema)
    exp = OrderedDict([("col_0", Array), ("col_1", List), ("col_2", Struct)])
    assert res == exp


def test_two_columns_to_dictionary_passes():
    inp_col_key_vals = [i for i in range(9)]
    inp_col_key = "inp_col_key"
    inp_col_val_vals = [i for i in range(1, 10)]
    inp_col_val = "inp_col_val"
    df = DataFrame({inp_col_key: inp_col_key_vals, inp_col_val: inp_col_val_vals})
    exp = dict(zip(inp_col_key_vals, inp_col_val_vals))
    res = two_columns_to_dictionary(
        df=df, key_col_name=inp_col_key, value_col_name=inp_col_val
    )
    assert res == exp


def test_two_columns_to_dictionary_raises_error_on_nested_keys():
    inp_col_key_vals = [[123], [234]]
    inp_col_key = "inp_col_key"
    inp_col_val_vals = [i for i in range(2)]
    inp_col_val = "inp_col_val"
    df = DataFrame({inp_col_key: inp_col_key_vals, inp_col_val: inp_col_val_vals})
    with pytest.raises(ValueError):
        two_columns_to_dictionary(
            df=df, key_col_name=inp_col_key, value_col_name=inp_col_val
        )


@pytest.fixture()
def duplicate_keys_df() -> DataFrame:
    inp_col_key_vals = ["duplicate", "duplicate", "not duplicate"]
    inp_col_key = "inp_col_key"
    inp_col_val_vals = [i for i in range(3)]
    inp_col_val = "inp_col_val"
    return DataFrame({inp_col_key: inp_col_key_vals, inp_col_val: inp_col_val_vals})


def test_two_columns_to_dictionary_raises_error_on_duplicate_keys_by_default(
    duplicate_keys_df: DataFrame,
):
    with pytest.raises(ValueError):
        two_columns_to_dictionary(
            df=duplicate_keys_df,
            key_col_name="inp_col_key",
            value_col_name="inp_col_val",
            allow_duplicates_keys=False,
        )


def test_two_columns_to_dictionary_allows_duplicate_keys(duplicate_keys_df: DataFrame):
    exp = {"duplicate": 1, "not duplicate": 2}
    res = two_columns_to_dictionary(
        df=duplicate_keys_df,
        key_col_name="inp_col_key",
        value_col_name="inp_col_val",
        allow_duplicates_keys=True,
    )
    assert res == exp



import pytest
from harley.transformations import snake_case_column_names, flatten_struct
from polars import DataFrame, LazyFrame, Series
from typing import Union
from harley.utils import polars_frames
from polars.testing import assert_frame_equal
from typing import List

column_names = [
    ["CamelCase", "camel_case"],
    ["This is Human case.", "this_is_human_case"],
    ["   This is preceding white space case.", "this_is_preceding_white_space_case"],
    ["This is succeeding white space case   ", "this_is_succeeding_white_space_case"],
    ["this.would.confuse.a.sql.query", "this_would_confuse_a_sql_query"],
    ["MixedUP CamelCase, with some Spaces", "mixed_up_camel_case_with_some_spaces"],
    [
        "mixed_up_ snake_case with some _spaces",
        "mixed_up_snake_case_with_some_spaces",
    ],
    ["kebab-case", "kebab_case"],
    ["SHOUTY_SNAKE_CASE", "shouty_snake_case"],
    ["snake_case", "snake_case"],
    ["ABcDE", "a_bc_de"],
    ["abcDEF", "abc_def"],
    ["ABC123dEEf456FOO", "abc123d_e_ef456_foo"],
    [
        "this-contains_ ALLKinds OfWord_Boundaries",
        "this_contains_all_kinds_of_word_boundaries",
    ],
    ["XΣXΣ baﬄe", "xσxς_baﬄe"],
    ["XMLHttpRequest", "xml_http_request"],
    ["FIELD_NAME11", "field_name11"],
    ["99BOTTLES", "99bottles"],
    ["FieldNamE11", "field_nam_e11"],
    ["abc123def456", "abc123def456"],
    ["abc123DEF456", "abc123_def456"],
    ["abc123Def456", "abc123_def456"],
    ["abc123DEf456", "abc123_d_ef456"],
    ["ABC123def456", "abc123def456"],
    ["ABC123DEF456", "abc123def456"],
    ["ABC123Def456", "abc123_def456"],
    ["ABC123DEf456", "abc123d_ef456"],
]


@pytest.mark.parametrize("frame_type", polars_frames)
@pytest.mark.parametrize(
    "input_column, exp",
    column_names,
)
def test_snake_case_column_names(
    frame_type: Union[LazyFrame, DataFrame], input_column: str, exp: str
):
    polars_frame = frame_type(data={input_column: range(10)})
    res = snake_case_column_names(polars_frame).columns
    assert res == [exp]


@pytest.fixture()
def data_with_struct() -> List[Series]:
    data = [
        Series(
            "ratings",
            [
                {"movie": "Cars", "theatre": "NE", "avg_rating": 4.5},
                {"movie": "Toy Story", "theatre": "ME", "avg_rating": 4.9},
            ],
        ),
        Series("name", ["George", "Yi Fong"]),
    ]
    return data


@pytest.mark.parametrize("frame_type", polars_frames)
@pytest.mark.timeout(1)
def test_flatten_struct(
    frame_type: Union[DataFrame, LazyFrame], data_with_struct: List[Series]
):
    inp = frame_type(data_with_struct)
    exp = DataFrame(
        [
            Series("name", ["George", "Yi Fong"]),
            Series("ratings:movie", ["Cars", "Toy Story"]),
            Series("ratings:theatre", ["NE", "ME"]),
            Series("ratings:avg_rating", [4.5, 4.9]),
        ]
    )
    if isinstance(res := flatten_struct(inp, "ratings"), LazyFrame):
        res = res.collect()
    assert_frame_equal(res, exp)


@pytest.mark.parametrize("frame_type", polars_frames)
@pytest.mark.timeout(1)
def test_flatten_struct_do_not_drop(
    frame_type: Union[DataFrame, LazyFrame], data_with_struct: List[Series]
):
    inp = frame_type(data_with_struct)
    exp = DataFrame(
        [
            Series("name", ["George", "Yi Fong"]),
            Series("ratings:movie", ["Cars", "Toy Story"]),
            Series("ratings:theatre", ["NE", "ME"]),
            Series("ratings:avg_rating", [4.5, 4.9]),
            Series(
                "ratings",
                [
                    {"movie": "Cars", "theatre": "NE", "avg_rating": 4.5},
                    {"movie": "Toy Story", "theatre": "ME", "avg_rating": 4.9},
                ],
            ),
        ]
    )
    if isinstance(
        res := flatten_struct(inp, "ratings", drop_original_struct=False), LazyFrame
    ):
        res = res.collect()
    assert_frame_equal(res, exp, check_column_order=False)


@pytest.mark.parametrize("frame_type", polars_frames)
@pytest.mark.timeout(1)
def test_flatten_struct_separator(
    frame_type: Union[DataFrame, LazyFrame], data_with_struct: List[Series]
):
    inp = frame_type(data_with_struct)
    exp = DataFrame(
        [
            Series("name", ["George", "Yi Fong"]),
            Series("ratings_movie", ["Cars", "Toy Story"]),
            Series("ratings_theatre", ["NE", "ME"]),
            Series("ratings_avg_rating", [4.5, 4.9]),
        ]
    )
    if isinstance(res := flatten_struct(inp, "ratings", separator="_"), LazyFrame):
        res = res.collect()
    assert_frame_equal(res, exp)


@pytest.fixture()
def nested_struct_data() -> dict:
    data = {
        "coords": [
            {"x": {"z": 1, "a": {"b": 2}}, "y": 4},
            {"x": {"z": 3, "a": {"b": 4}}, "y": 9},
            {"x": {"z": 5, "a": {"b": 6}}, "y": 16},
        ],
        "multiply": [10, 2, 3],
    }
    return data


@pytest.mark.parametrize("frame_type", polars_frames)
@pytest.mark.timeout(1)
def test_flatten_struct_recursive(
    frame_type: Union[DataFrame, LazyFrame], nested_struct_data: List[Series]
):
    inp = frame_type(nested_struct_data)
    exp = DataFrame(
        [
            Series("coords:x:z", [1, 3, 5]),
            Series("coords:x:a:b", [2, 4, 6]),
            Series("coords:y", [4, 9, 16]),
            Series("multiply", [10, 2, 3]),
        ]
    )
    if isinstance(
        res := flatten_struct(df=inp, struct_columns="coords", recursive=True),
        LazyFrame,
    ):
        res = res.collect()
    assert_frame_equal(res, exp, check_column_order=False)


@pytest.mark.parametrize("frame_type", polars_frames)
@pytest.mark.timeout(1)
def test_flatten_struct_recursive_limit(
    frame_type: Union[DataFrame, LazyFrame], nested_struct_data: List[Series]
):
    inp = frame_type(nested_struct_data)
    exp = DataFrame(
        [
            Series("coords:x:z", [1, 3, 5]),
            Series("coords:x:a", [{"b": 2}, {"b": 4}, {"b": 6}]),
            Series("coords:y", [4, 9, 16]),
            Series("multiply", [10, 2, 3]),
        ]
    )
    if isinstance(
        res := flatten_struct(df=inp, struct_columns="coords", recursive=True, limit=2),
        LazyFrame,
    ):
        res = res.collect()
    assert_frame_equal(res, exp, check_column_order=False)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generate the code reference pages and navigation.

Script was taken from
https://mkdocstrings.github.io/recipes/#automatic-code-reference-pages
"""

from pathlib import Path

import mkdocs_gen_files

nav = mkdocs_gen_files.Nav()

for path in sorted(Path(".").rglob("harley/**/*.py")):
    if "utils" in str(path):
        continue
    module_path = path.relative_to(".").with_suffix("")
    doc_path = path.relative_to(".").with_suffix(".md")
    full_doc_path = Path("reference", doc_path)

    parts = tuple(module_path.parts)

    if parts[-1] == "__init__":
        parts = parts[:-1]
        doc_path = doc_path.with_name("index.md")
        full_doc_path = full_doc_path.with_name("index.md")
    elif parts[-1] == "__main__":
        continue

    nav[parts] = doc_path.as_posix()  #

    with mkdocs_gen_files.open(full_doc_path, "w") as fd:
        ident = ".".join(parts)
        fd.write(f"::: {ident}")

    mkdocs_gen_files.set_edit_path(full_doc_path, path)

with mkdocs_gen_files.open("reference/SUMMARY.md", "w") as nav_file:
    nav_file.writelines(nav.build_literate_nav())


from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Union

import polars as pl

from harley.utils import parse_into_expr, register_plugin, parse_version

if TYPE_CHECKING:
    from polars.type_aliases import IntoExpr

if parse_version(pl.__version__) < parse_version("0.20.16"):
    from polars.utils.udfs import _get_shared_lib_location

    lib: str | Path = _get_shared_lib_location(__file__)
else:
    lib = Path(__file__).parent


def div_or_else(
    dividend: IntoExpr, divisor: IntoExpr, or_else: Union[int, float] = 0.0
) -> IntoExpr:
    """
    Returns the result of dividing one expression by another, with an optional default
    value if the divisor is zero.

    :param dividend: The value that will be divided
    :type dividend: IntoExpr
    :param divisor: The value by which
    the `dividend` will be divided. If the `divisor` is zero, the function will return the `or_else`
    value instead of performing the division
    :type divisor: IntoExpr
    :param or_else: TA default value that will
    be returned if the divisor is zero. It is a numeric value (either an integer or a float) and is set
    to 0.0 by default if not provided explicitly.
    :type or_else: Union[int, float]
    :return: The result of the division of `dividend` by `divisor`,
    or the default value `or_else` if the divisor is zero.
    """
    dividend = parse_into_expr(dividend)
    divisor = parse_into_expr(divisor)
    return register_plugin(
        args=[dividend, divisor],
        symbol="div_or_else",
        is_elementwise=True,
        kwargs={"or_else": float(or_else)},
        lib=lib,
    )



import pytest
from harley.utils import polars_frames
from harley.to_boolean import is_null_or_blank, is_true, is_truthy, is_false, is_falsey
from polars import DataFrame, LazyFrame
from polars.testing import assert_frame_equal
from typing import List, Callable, Union


@pytest.mark.parametrize("frame_type", polars_frames)
@pytest.mark.parametrize(
    "inp, exp, all_white_space_as_null",
    [
        [[" ", "", "not blank", None], [False, True, False, True], False],
        [[" ", "", "not blank", None], [True, True, False, True], True],
        [[[], [123], None], [True, False, True], False],
    ],
)
def test_is_null_or_blank(
    frame_type: Union[DataFrame, LazyFrame],
    inp: List,
    exp: List,
    all_white_space_as_null: bool,
):
    data = frame_type({"value": inp})
    if isinstance(
        res := data.select(res=is_null_or_blank("value", all_white_space_as_null)),
        LazyFrame,
    ):
        res = res.collect()
    exp = DataFrame({"res": exp})
    assert_frame_equal(res, exp)


@pytest.mark.parametrize("frame_type", polars_frames)
@pytest.mark.parametrize("tested_function", [is_truthy, is_true])
def test_is_true_truthy(
    frame_type: Union[DataFrame, LazyFrame], tested_function: Callable
):
    inp = [True, False, None]
    exp = DataFrame({"res": [True, False, False]})
    data = frame_type({"value": inp})
    if isinstance(res := data.select(res=tested_function("value")), LazyFrame):
        res = res.collect()
    assert_frame_equal(res, exp)


@pytest.mark.parametrize("frame_type", polars_frames)
def test_is_false(frame_type: Union[DataFrame, LazyFrame]):
    inp = [True, False, None]
    exp = DataFrame({"res": [False, True, False]})
    data = frame_type({"value": inp})
    if isinstance(res := data.select(res=is_false("value")), LazyFrame):
        res = res.collect()
    assert_frame_equal(res, exp)


@pytest.mark.parametrize("frame_type", polars_frames)
def test_is_falsey(frame_type: Union[DataFrame, LazyFrame]):
    inp = [True, False, None]
    exp = DataFrame({"res": [False, True, True]})
    data = frame_type({"value": inp})
    if isinstance(res := data.select(res=is_falsey("value")), LazyFrame):
        res = res.collect()
    assert_frame_equal(res, exp)
